{{
******************************************
*            quick_mult.spin             *
* (c) Copyright 2011 Bueno Systems, Inc. *
*    See end of file for terms of use.   *
******************************************

Program to produce efficient PASM multiply routines, given
a semi-constant multiplier. As a demonstration, it produces
ten multiplication routines with different multipliers, and
tests each with ten multiplicands. The Spin and quick_mult
products are displayed for comparison.

}}

CON

  _clkmode      = xtal1 + pll16x
  _xinfreq      = 5_000_000

  NEW           = 1
  MPY           = 2

OBJ

  pst   :       "Parallax Serial Terminal"
  rnd   :       "RealRandom"

VAR

  long command, arg, mult[34]

PUB start | mplr, mpcd, n, i, rand, product

  rnd.start
  rand := rnd.random
  rnd.stop
  pst.start(38400) 
  cognew(@quick_mult, @command)
  repeat 10
    mplr := ?rand ~> 18
    n := new_multiplier(mplr)
    i := n >> 8
    n &= $ff
    pst.str(string("Multiplier = "))
    pst.bin(mplr, 32)
    pst.str(string(" = "))
    pst.dec(mplr)
    pst.char(13)
    pst.str(string("Instruction cycles to produce = "))
    pst.dec(i)
    pst.str(string(13, "Instructions in mult routine  = "))
    pst.dec(n)
    pst.char(13)
    pst.char(13)
    repeat i from 0 to n - 1
      pst.bin(mult[i] >> 26, 6)
      pst.char(" ")
      pst.bin(mult[i] >> 22 & $f, 4)
      pst.char(" ") 
      pst.bin(mult[i] >> 18 & $f, 4)
      pst.char(" ") 
      pst.bin(mult[i] >> 9 & $1ff, 9)
      pst.char(" ") 
      pst.bin(mult[i] & $1ff, 9)
      case mult[i] >> 26
        %101000: pst.str(string("    mov   :product,:mpcd", 13))
        %101001: pst.str(string("    neg   :product,:mpcd", 13))
        %100000: pst.str(string("    add   :product,:mpcd", 13))
        %100001: pst.str(string("    sub   :product,:mpcd", 13))
        %010111: pst.str(string("    ret", 13))
        %001011: pst.str(string("    shl   :mpcd,#"))
                 pst.dec(mult[i] & $1f)
                 pst.char(13)
        other:   pst.str(string("**ERROR**", 13))
    pst.char(13)
    repeat 10
      mpcd := ?rand ~> 18
      pst.dec(mplr)
      pst.str(string(" * "))
      pst.dec(mpcd)
      pst.str(string(" = "))
      pst.dec(mplr * mpcd)
      pst.str(string(" = "))
      pst.dec(multiply(mpcd))
      pst.char(13)
    pst.str(string("-------", 13))
     
PUB new_multiplier(multiplier)

'' Produce a new multiply routine, given a new multiplier.

  arg := multiplier
  command := NEW
  repeat while command
  return arg

PUB multiply(multiplicand)

'' Compute a product from a previously-supplied multiplier and the
'' given multiplicand.

  arg := multiplicand
  command := MPY
  repeat while command
  return arg

DAT

              org       0

quick_mult    mov       :argaddr,par            'Get the argument address.
              add       :argaddr,#4

:cmd_lp       rdlong    :cmd,par wz             'Read a command. What is it?
        if_z  jmp       #:cmd_lp                '0: Try again.

              djnz      :cmd,#:do_mpy           '1: Do the multiply.

'=======[ Build the multiply routine. ]========================================

:make_mult    rdlong    :bin,:argaddr           '2: Build routine, so get the multiplier.
              neg       :time,cnt
              neg       :csdn,:bin              'Convert to CSD form.
              jmpret    :bin2csd_ret,#:bin2csd
              add       :csdn,:csdp

              neg       :i,#1                   'Initialize shift counter.
              mov       :c,#0                   'Initialize "did mov/neg" flag.
              movd      :putinstr,#:mult+1      'Initialize instruction pointer.

:build_lp     or        :csdp,:csdn nr,wz       'Any "one" bits left?
        if_z  jmp       #:build_done            '  No:  All done.

:inner_lp     test      :csdp,#1 wz             'Set z from :csdp LSB. 
              shr       :csdp,#1
              shr       :csdn,#1 wc             'Set c from :csdn LSB.
              add       :i,#1                   'Increment shift counter,
  if_z_and_nc jmp       #:inner_lp               'Both bits zero, so try next pair.

              tjz       :i,#:noshift            'Shift count non-zero?

              mov       :instr,:shl_instr       '  Yes: Insert shift instruction.
              add       :instr,:i
              jmpret    :putinstr_ret,#:putinstr
              mov       :i,#0                   '       Reinitialize shift count.

:noshift      tjz       :c,#:do_mov             'First instruction?

        if_nz mov       :instr,:add_instr       '  No:  Positive: Insert add.
        if_c  mov       :instr,:sub_instr       '       Negative: Insert sub.
              jmp       #:putit                 '       Go do it.

:do_mov if_nz mov       :instr,:mov_instr       '  Yes: Positive: Insert mov.
        if_c  mov       :instr,:neg_instr       '       Negative: Insert neg.
              mov       :c,#1                   '       Set "got first instruction" flag.

:putit        jmpret    :putinstr_ret,#:putinstr 'Put the instruction in the routine.
              jmp       #:build_lp              'Back to continue building.        

:build_done   mov       :instr,:ret_instr       'Done building: Insert return instruction.
              jmpret    :putinstr_ret,#:putinstr
              add       :time,cnt
              mov       :i,:putinstr            'Return length of routine created.
              shr       :i,#9
              and       :i,#$1ff
              sub       :i,#:mult+1
              shr       :time,#2
              shl       :time,#8
              add       :i,:time
              wrlong    :i,:argaddr
              and       :i,#$ff
              jmpret    :dump_ret,#:dump        'Copy the constructed routine to hub.
              wrlong    :zero,par               'Signal completion.
              jmp       #:cmd_lp                'Back for next command.

'-------[ Insert an instruction into the multiply routine. ]-------------------

:putinstr     mov       0-0,:instr              'Insert instruction in mult routine.
              add       :putinstr,:inc_dest     'Increment insertion pointer.
:putinstr_ret ret                               'Return.

'-------[ Convert binary to CSD format. ]--------------------------------------

:bin2csd      mov       :csdp,#0                'Initialize positive result.
              mov       :c,#0                   'Initialize intermediate carry.
              mov       :i,#32                  'Initialize bit counter.

:conv_lp      test      :bin,#%01 wz            'Test lower bits of number.
              test      :bin,#%10 wc
 if_nz_and_c  mov       :c,#1                   '%11: Set intermediate carry.
        if_c  andn      :bin,#1                 '%1*: Clear LSB of number.
        if_nc xor       :bin,:c                 '%0*: XOR LSB of number with intermediate carry.
 if_nc_and_z  mov       :c,#0                   '%00: Clear intermediate carry.
              shr       :bin,#1 wc              'Shift new LSB of value into...
              rcr       :csdp,#1                '  ...MSB of positive result.
              djnz      :i,#:conv_lp            'Back for another bit.
              
:bin2csd_ret  ret

'-------[ Copy multiply routine to hub for inspection. ]-----------------------

:dump         movd      :wr_hub,#:mult+1

:dump_lp      add       :argaddr,#4
:wr_hub       wrlong    0-0,:argaddr
              add       :wr_hub,:inc_dest
              djnz      :i,#:dump_lp

              mov       :argaddr,par
              add       :argaddr,#4
:dump_ret     ret

'=======[ Perform the multiplication. ]----------------------------------------

:do_mpy       rdlong    :mpcd,:argaddr          'Get the multiplicand.
              jmpret    :mult,#:mult+1          'Call the multiply routine.
              wrlong    :product,:argaddr       'Write the product to the hub.
              wrlong    :zero,par               'Signal completion.
              jmp       #:cmd_lp                'Back for another command.

'-------[ Constructed multiply routine ]---------------------------------------

:mult         jmp       #0-0                    'Return vector.
              mov       :product,#0             'Dummy routine.
              jmp       :mult                   'Return.
              long      0-0[31]                 'More buffer space for routine.

'-------[ List of possible instructions ]--------------------------------------

:mov_instr    mov       :product,:mpcd
:neg_instr    neg       :product,:mpcd     
:add_instr    add       :product,:mpcd
:sub_instr    sub       :product,:mpcd
:shl_instr    shl       :mpcd,#0-0
:ret_instr    jmp       :mult

'-------[ Constants ]----------------------------------------------------------

:zero         long      0
:inc_dest     long      1 << 9

'-------[ Variables ]----------------------------------------------------------

:argaddr      res       1
:cmd          res       1
:c            res       1
:csdp         res       1
:csdn         res       1
:bin          res       1
:i            res       1
:mpcd         res       1
:mplr         res       1
:product      res       1
:instr        res       1
:time         res       1

''=======[ License ]==========================================================
{{{
***************************************************************************************
*                                                                                     *
* Permission is hereby granted, free of charge, to any person obtaining a copy of     *
* this software and associated documentation files (the "Software"), to deal in       *
* the Software without restriction, including without limitation the rights to use,   *
* copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the     *
* Software, and to permit persons to whom the Software is furnished to do so,         *
* subject to the following conditions:                                                *
*                                                                                     *
* The above copyright notice and this permission notice shall be included in all      *
* copies or substantial portions of the Software.                                     *
*                                                                                     *
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, *
* INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A       *
* PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT  *
* HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION   *
* OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE      *
* SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.                              *
*                                                                                     *
***************************************************************************************
}}
